class Pask
  require "thread"
  require_relative "pask/event"

  NotStartedError  = Class.new(RuntimeError)
  NotFinishedError = Class.new(RuntimeError)

  VERSION          = "0.1.0".freeze
  RUN_STATE        = "run".freeze
  SLEEP_STATE      = "sleep".freeze
  TERMINATED_STATE = [nil, false].freeze
  CATCH_ALL        = Proc.new { true }

  def self.version
    VERSION
  end

  def initialize
    @thread = nil
    @queue = nil
    @predicate = CATCH_ALL
    @predicate_copy = nil
    @block = nil
    @block_copy = nil
  end

  def interested?(callable = nil, &block)
    predicate = callable || block
    unless predicate
      raise ArgumentError, "no predicate given to #{__method__}()"
    end
    @predicate = predicate
  end

  def started?
    if @queue and @thread
      true
    else
      false
    end
  end

  def running?
    if @thread and @thread.status == RUN_STATE
      true
    else
      false
    end
  end

  def finished?
    if @thread and TERMINATED_STATE.include?(@thread.status)
      true
    else
      false
    end
  end

  def sleeping?
    if @thread and @thread.status == SLEEP_STATE
      true
    else
      false
    end
  end

  def resume
    unless started?
      raise NotStartedError, "tracer has not been started"
    end
    if sleeping?
      @thread.wakeup
      @queue.deq
    end
  end

  def start
    if started? and not finished?
      raise NotFinishedError, "tracer has not finished"
    end
    @queue = Queue.new
    @thread = Thread.new do
      @predicate_copy, @block_copy = @predicate, @block
      Thread.current.set_trace_func method(:on_event).to_proc
      @block_copy.call
      Thread.current.set_trace_func(nil)
      @queue.enq nil
    end
    @queue.deq
  end

  def trace(&block)
    @block = block
  end

  def consume!
    events = [start]
    current_event = nil
    events.push(current_event) while current_event = resume
    events
  end

private
  def on_event(name, file, lineno, method, binding, _)
    event = Event.new name, file: file, method: method, lineno: lineno, binding: binding
    if event.file != __FILE__ and @predicate_copy.call(event)
      @queue.enq(event)
      Thread.stop
    end
  rescue Exception => e
    warn "TRACER CRASHED"
    Thread.current.set_trace_func(nil)
  end
end
